/*-
 * Copyright (C) @BABOLO  2009 Apr 27 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *  
 *  For Yandex spamooborona details please visit
 *  http://so.yandex.ru/
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2009 Apr 27 http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: spamooborona1024.c,v 1.1.2.2 2009/05/05 22:41:09 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sysexits.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include <zlib.h>
#include <err.h>
#include <mife.h>
#include <babolo/parser.h>
#include <babolo/BLINflag.h>

#include "mh.h"
#include "so.h"

/* ,  spamooborona1024_query()   
 *     (?) 
 */
#define SO1024MXMS      65536
#define SO1024ANSS      4096
#define SO1024HOST      "213.180.204.68"    /* so-plus.yandex.ru */
#define SO1024PORT      "2525"
#define SO1024A_OK      "OK"
#define SPAMOOBORONA1024_BRJC 0x000001 /* REJECT                                                     */
#define SPAMOOBORONA1024_BSPM 0x000002 /* SPAM                                                       */
#define SPAMOOBORONA1024_BSPL 0x000004 /*                                                 */
#define SPAMOOBORONA1024_BMSK (SPAMOOBORONA1024_BRJC | SPAMOOBORONA1024_BSPM | SPAMOOBORONA1024_BSPL)
#define SPAMOOBORONA1024_RRDY 0x001000 /*                       */
#define SPAMOOBORONA1024_TEST 0x002000 /*                            */
#define SPAMOOBORONA1024_RDNS 0x004000 /*  PTR  IP ,    FQDN         */
#define SO1024_EFLG           0x008000 /* -E flag                                                    */
#define SPAMOOBORONA1024_QSFT 16       /*                                           */
#define SPAMOOBORONA1024_QRJC (SPAMOOBORONA1024_BRJC << SPAMOOBORONA1024_QSFT)
                                       /*   REJECT                               */
#define SPAMOOBORONA1024_QSPM (SPAMOOBORONA1024_BSPM << SPAMOOBORONA1024_QSFT)
                                       /*   SPAM                                 */
#define SPAMOOBORONA1024_QSPL (SPAMOOBORONA1024_BSPL << SPAMOOBORONA1024_QSFT)
                                       /*                            */
#define SPAMOOBORONA1024_QMSK (SPAMOOBORONA1024_BMSK << SPAMOOBORONA1024_QSFT)
#define SPAMOOBORONA1024_ASFT 20       /*                                            */
#define SPAMOOBORONA1024_ARJC (SPAMOOBORONA1024_BRJC << SPAMOOBORONA1024_ASFT)
                                       /* REJECT                                             */
#define SPAMOOBORONA1024_ASPM (SPAMOOBORONA1024_BSPM << SPAMOOBORONA1024_ASFT)
                                       /* SPAM                                               */
#define SPAMOOBORONA1024_ASPL (SPAMOOBORONA1024_BSPL << SPAMOOBORONA1024_ASFT)
                                       /*                                         */
#define SPAMOOBORONA1024_AMSK (SPAMOOBORONA1024_BMSK << SPAMOOBORONA1024_ASFT)
#define SO1024_MDNS           0x800000 /* free() FQDN                                                */

typedef struct spamooborona1024_root {
    u_int32_t flags;       /*                                                                   */
    float     spamlevel;   /*                                       */
    int       s;           /* socket                                             */
    int       asld;        /*   --                           */
    size_t    maxmailsize; /*                             */
    char     *host;        /* IP                                            */
    char     *port;        /*                                                 */
    FILE     *out;         /*                                                            */
    mife_descriptor *md;
} spamooborona1024_root;

static char testan[] =
"SODAEMON 127\n"
"REJECT 0\n"
"REJECTSTR \n"
"SPAM 1\n"
"SPAMSTR [SPAM 106.20]\n"
"X-Spam-Ystatus: hits=106.2\n"
"X-Spam-Flag: YES\n"
"X-Spam-Yversion: Spamooborona-3.0\n"
;

static int
so1024_cmd(spamooborona1024_root *sprp, char *cmd, ...) {
    int len, ex = EX_OK;
    char *cmdval;
    va_list ap;

    va_start(ap, cmd);
    if  ((len = vasprintf(&cmdval, cmd, ap)) < 0) {
        ifBLIN_QV1(sprp->flags) warn("asprintf #2 %s", cmd);
        if  (errno == ENOMEM) ex = EX_TEMPFAIL;
        else if (errno == EILSEQ) ex = EX_DATAERR;
        else ex = EX_OSERR;
    } else {
        char answ[4];
        bzero(answ, 4 * sizeof(char));
        if  (sprp->flags & SPAMOOBORONA1024_TEST) {
            ifBLIN_QV1(sprp->flags) fprintf(stderr, "cmd=%s~\n", cmdval);
        } else {
            ifBLIN_QV4(sprp->flags) fprintf(stderr, "cmd=%s~\n", cmdval);
            if  (send(sprp->s, cmdval, len, 0) < 0) {
                ifBLIN_QV1(sprp->flags) warn("send #1");
                ex = EX_IOERR;
        }   }
        if  (!ex && !(sprp->flags & SPAMOOBORONA1024_TEST)) {
            if  ((len = read(sprp->s, answ, strlen(SO1024A_OK) + 1)) < 0) {
                ifBLIN_QV1(sprp->flags) warn("read #1 %s", cmdval);
                ex = EX_IOERR;
            } else if (len < strlen(SO1024A_OK)) {
                ifBLIN_QV1(sprp->flags) warnx("read #1 proto(%d) %s", len, cmdval);
                ex = EX_PROTOCOL;
            } else if (strncmp(answ, SO1024A_OK, strlen(SO1024A_OK))) {
                ifBLIN_QV1(sprp->flags) warnx("read #2 proto(%s) %s", answ, cmdval);
                ex = EX_PROTOCOL;
        }   }
        if  (  !ex
            && !(sprp->flags & SPAMOOBORONA1024_TEST)
            && strncmp(answ, SO1024A_OK, strlen(SO1024A_OK))
            ) {
            ifBLIN_QV1(sprp->flags) warnx("answer %s", answ);
            ex = EX_PROTOCOL;
        }
        free(cmdval);
    }
    return(ex);
}

static ssize_t
so1024_crlf(spamooborona1024_root *sprp, u_char *text, size_t len, int to) {
    enum classes {n0, nr, nl, no, nx};
    /* n0 - EOF
       nr - \r
       nl - \n
       no - 
       nx
     */
    static const enum classes class[257] =
    { n0
    , no,no,no,no, no,no,no,no, no,no,nl,no, no,nr,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no

    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no

    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no

    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    , no,no,no,no, no,no,no,no, no,no,no,no, no,no,no,no
    };

#   define C_nl 0x800000 /*  \r\n                                                             */
#   define C_cl 0x400000 /*  \r\n                                               */
#   define C_nm 0x200000 /*                                                    */
#   define C_nx 0x100000 /*                                                                 */
#   define C_to 0x080000 /*                                                  */
#   define C_state    0xFF /*                                                      */

    enum states {d0, d1, dr, br, bn, dx};
    /* d0 -   
       d1 -  
       dr -  \r
       br -  \r  
       dx -  
     */

    static const u_int32_t automa[dx][nx] =
    /*           n0            nr                 nl                 no   */
    {{           dx,      C_nx|br, C_cl|C_to|C_nx|d0,      C_nm|C_nx|d1} /* d0 */
    ,{ C_nl|     dx,      C_nx|dr, C_nl|     C_nx|d0,      C_nm|C_nx|d1} /* d1 */
    ,{ C_nl|     dx,      C_nx|br, C_nl|     C_nx|d0, C_nl|C_nm|C_nx|d1} /* dr */
    ,{ C_cl|C_to|dx, C_cl|C_to|d0, C_cl|C_to|C_nx|d0, C_cl|C_to|     d0} /* br */
    };

#if DEBUG
    static const char *clasnames[] = {"n0", "nr", "nl", "no", "nx"};
    static const char *pn[] = 
    { "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
    , "nl", "cl", "nm", "nx", "to", "- ", "- ", "- "
    , "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
    };
    static const char *statenames[] = {"d0", "d1", "dr", "br", "bn", "dx"};
#endif

    u_int32_t state, control, mask;
    int ex = EX_OK, c;
    size_t p;

    if  (to) mask = ~C_cl; else mask = ~C_to;
    ifBLIN_QV5(sprp->flags) BLIN_STATEHEADER(stderr, pn)
    for (state = 0, p = 0; state < dx;) {
        if  (p >= len) c = -1; else c = text[p];
        control = automa[state][class[c + 1]] & mask;
        ifBLIN_QV5(sprp->flags) BLIN_STATEBODY( stderr\
                                              , pn\
                                              , (c & 0xFF)\
                                              , clasnames[class[c + 1]]\
                                              , text\
                                              , p\
                                              , BLIN_I(text)\
                                              , control\
                                              , statenames\
                                              , state\
                                              , C_state\
                                              , (text[p + BLIN_STATEVAR_i]>=' ')\
                                              )
        if  ((control & (C_nl | C_cl)) && ((ex = fprintf(sprp->out, "\r\n")) < 0)) break;
        if  ((control & C_nm)  && ((ex = putc(c, sprp->out)) < 0)) break;
        if  (control & C_nx) p++;
        if  (control & C_to) break;
        state = control & C_state;
    }
    if  (ex < 0) ex = -EX_IOERR; else ex = p;
    return(ex);
}

int
spamooborona1024_query( spamooborona1024_root *sprp
                      , char   *h_ip
                      , char   *h_fqdn
                      , char   *helo
                      , char   *mailfrom
                      , char  **rcpto
                      , ssize_t mailsize
                      , char   *maildata
                      ) {
    int ex, hlen = 0;
    char *body1;
    uLongf csize = 0;
    u_char *cdata = NULL;
    static const struct addrinfo hint =
        {AI_ADDRCONFIG, PF_UNSPEC, SOCK_STREAM, IPPROTO_TCP, 0, NULL, NULL, NULL};

    /**************************************************************************************************
     * ,       
     */
    ex = EX_OK;
    sprp->flags &= ~(SO1024_MDNS | SPAMOOBORONA1024_AMSK);
    if  (sprp->maxmailsize > SO1024MXMS) sprp->maxmailsize = SO1024MXMS;
    if  (mailsize > sprp->maxmailsize) mailsize = sprp->maxmailsize;
    ifBLIN_QV4(sprp->flags) {
        size_t i;
        fprintf( stderr, "+spamooborona1024_query %08X lv=%f max=%d sock=%d sld=%d\n"
               , sprp->flags, sprp->spamlevel, sprp->maxmailsize, sprp->s, sprp->asld
               );
        if  (sprp->host) fprintf(stderr, " host=%s~", sprp->host);
        if  (sprp->port) fprintf(stderr, " port=%s~", sprp->port);
        if  (h_ip) fprintf(stderr, " CLIENT=%s~", h_ip);
        if  (h_fqdn) fprintf(stderr, " FQDN=%s~", h_fqdn);
        if  (mailfrom) fprintf(stderr, " SENDER=%s~", mailfrom);
        fprintf(stderr, "\n");
        if  (sprp) {
            char *c;

            fprintf(stderr, " RECIPIENTS");
            c = "";
            for (i = 0; rcpto[i]; i++) {
                fprintf(stderr, "%s=%s~\n", c, rcpto[i]);
                c = "           ";
            }
            fprintf(stderr, "\n");
    }   }
    ifBLIN_QV6(sprp->flags) babolo_treedump(&sowords);

    /**************************************************************************************************
     *   ,
     *  ,   ,
     *       
     */
    if  (!ex) {
#ifdef compressBound
        csize = compressBound(mailsize);
#else
        csize = mailsize + (mailsize >> 12) + (mailsize >> 14) + 11;
#endif
        if  (!(cdata = malloc(csize))) {
            ifBLIN_QV1(sprp->flags) warnx("No memory spamooborona1024_query #1 of %d", (int)csize);
            ex = EX_OSERR;
    }   }
    if  (!ex) {
        ex = compress((Bytef*)cdata, &csize, (Bytef*)maildata, (uLong)mailsize);
        if  (ex >= 0) {
            ex = 0;
        } else if (ex == Z_MEM_ERROR) {
            ifBLIN_QV1(sprp->flags) warnx("spamooborona1024_query Z_MEM_ERROR");
            ex = EX_OSERR;
        } else if (ex == Z_BUF_ERROR) {
            ifBLIN_QV1(sprp->flags) warnx("spamooborona1024_query Z_BUF_ERROR");
            ex = EX_SOFTWARE;
        } else {
            ifBLIN_QV1(sprp->flags) warnx("spamooborona1024_query %d", ex);
            ex = EX_SOFTWARE;
    }   }

    /**************************************************************************************************
     *    SMTP ,
     *   ,    .
     *      
     */
    if  (!ex) {
        struct addrinfo *ai = NULL;
        int her;

        if  ((sprp->flags & SPAMOOBORONA1024_RDNS) && h_ip && !h_fqdn) {
        ;   if  ((her = getaddrinfo(h_ip, "0", &hint, &ai))) {
        ;       ifBLIN_QV2(sprp->flags) warnx("getaddrinfo for SMTP server: %s", gai_strerror(her));
        ;   } else {
                struct addrinfo *ap;

        ;       for (ap = ai; ap; ap = ap->ai_next) {
        ;           ifBLIN_QV5(sprp->flags) {
                        size_t i;
        ;               fprintf( stderr
                               , " C flags=%08X family=%d socktype=%d proto=%d addrlen=%d addr="
                               , ap->ai_flags
                               , ap->ai_family
                               , ap->ai_socktype
                               , ap->ai_protocol
                               , ap->ai_addrlen
                               );
        ;               for (i = 0; i < ap->ai_addrlen; i++)
                            fprintf(stderr, "%02X", ((u_char*)(ap->ai_addr))[i]);
        ;               fprintf(stderr, "\n");
        ;           }
        ;           if  (!(h_fqdn = malloc(NI_MAXHOST + 1))) {
        ;               ifBLIN_QV1(sprp->flags)
                            warnx("No memory spamooborona1024_query #2 of %d", NI_MAXHOST + 1);
        ;               ex = EX_OSERR;
        ;               break;
        ;           }
        ;           sprp->flags |= SO1024_MDNS;
        ;           if  ((her = getnameinfo( ap->ai_addr
                                           , ap->ai_addrlen
                                           , h_fqdn
                                           , NI_MAXHOST
                                           , NULL
                                           , 0
                                           , NI_NAMEREQD
                        ))                 ) {
        ;               ifBLIN_QV2(sprp->flags)
                            warnx("getnameinfo for SMTP server: %s", gai_strerror(her));
        ;           } else {
        ;               break;
        ;       }   }
        ;       freeaddrinfo(ai);
        ;       ai = NULL;
    }   }   }

    /**************************************************************************************************
     *    ,
     *      
     */
    if  (!ex && !(sprp->flags & SPAMOOBORONA1024_RRDY)) {
        struct addrinfo *ai = NULL;
        int her;

        ex = EX_NOHOST;
        if  (!sprp->port) sprp->port = SO1024PORT;
        if  (!sprp->host) sprp->host = SO1024HOST;
        sprp->s = -1;
        if  ((her = getaddrinfo(sprp->host, sprp->port, &hint, &ai))) {
            ifBLIN_QV1(sprp->flags)
                warnx("getaddrinfo for spamooborona1024 server: %s", gai_strerror(her));
            ex = EX_NOHOST;
        } else {
            struct addrinfo *ap;

        ;   for (ap = ai; ap; ap = ap->ai_next) {
        ;       ifBLIN_QV5(sprp->flags) {
                    size_t i;
        ;           fprintf( stderr, " S flags=%08X family=%d socktype=%d proto=%d addrlen=%d addr="
                           , ap->ai_flags
                           , ap->ai_family
                           , ap->ai_socktype
                           , ap->ai_protocol
                           , ap->ai_addrlen
                           );
        ;           for (i = 0; i < ap->ai_addrlen; i++)
                        fprintf(stderr, "%02X", ((u_char*)(ap->ai_addr))[i]);
        ;           fprintf(stderr, "\n");
        ;       }
        ;       if  ((sprp->s = socket(ap->ai_family, ap->ai_socktype, ap->ai_protocol)) < 0) {
        ;           ifBLIN_QV3(sprp->flags)
                        fprintf( stderr, "In loop socket(%d, %d, %d)\n"
                               , ap->ai_family, ap->ai_socktype, ap->ai_protocol
                               );
        ;           continue;
        ;       }
        ;       if  (  !(sprp->flags & SPAMOOBORONA1024_TEST)
                    && ((ex = connect(sprp->s, ap->ai_addr, ap->ai_addrlen)) < 0)
                    ) {
        ;           ifBLIN_QV3(sprp->flags)
                        fprintf(stderr, "In loop connect(%d, *, %d)\n", sprp->s, ap->ai_addrlen);
        ;           close(sprp->s);
        ;           sprp->s = -1;
        ;           continue;
        ;       } else ex = EX_OK;
        ;       break;
        ;   }
        ;   freeaddrinfo(ai);
        ;   ai = NULL;
        ;   sprp->flags |= SPAMOOBORONA1024_RRDY;
    }   }

    /**************************************************************************************************
     *    
     */
    if  (!ex) ex = so1024_cmd(sprp, "CONNECT %s [%s]", h_fqdn ? h_fqdn : "", h_ip ? h_ip : "");
    if  (sprp->flags & SO1024_MDNS) {
        free(h_fqdn);
        h_fqdn = NULL;
        sprp->flags &= ~SO1024_MDNS;
    }
    if  (!ex && helo) ex = so1024_cmd(sprp, "HELO %s", helo);
    if  (!ex) ex = so1024_cmd(sprp, "MAILFROM %s", mailfrom ? mailfrom : "");
    if  (!ex && sprp) {
        size_t i;
        for (i = 0; rcpto[i]; i++) {
            ex = so1024_cmd(sprp, "RCPTTO %s", rcpto[i]);
            if  (ex) break;
    }   }
    if  (!ex) ex = so1024_cmd(sprp, "DATA SIZE=%d", csize);

    /**************************************************************************************************
     *     
     */
    if  (!ex) {
        if  (  ((sprp->flags & BLIN_VER1) && (sprp->flags & SPAMOOBORONA1024_TEST))
            || (sprp->flags & BLIN_VER4)
            ) {
            size_t i;
            for (i = 0; i < csize; i++) {
                if  (!(i & 15)) fprintf(stderr, " %08X", i);
                if  (!(i & 1)) fprintf(stderr, " ");
                if  (!(i & 7)) fprintf(stderr, " ");
                fprintf(stderr, " %02X", cdata[i]);
                if  ((i & 15) == 15) fprintf(stderr, "\n");
            }
            if  (i & 15) fprintf(stderr, "\n");
        }
        if  (!(sprp->flags & SPAMOOBORONA1024_TEST)) {
            ssize_t i;
            while (csize > 0) {
                if  ((i = send(sprp->s, cdata, csize, 0)) > 0) {
                    csize -= i;
                    cdata += i;
                } else if (errno != EINTR) {
                    ifBLIN_QV1(sprp->flags) warn("send #2");
                    ex = EX_IOERR;
                    break;
    }   }   }   }

    /**************************************************************************************************
     *     
     */
    if  (!ex) {
        if  (!(body1 = calloc(SO1024ANSS, sizeof(char)))) {
            ifBLIN_QV1(sprp->flags) warnx("No memory spamooborona1024_query #3 of %d", SO1024ANSS);
            ex = EX_OSERR;
    }   }
    if  (!ex) {
        ssize_t i;

        if  (sprp->flags & SPAMOOBORONA1024_TEST) {
            strncpy(body1, testan, SO1024ANSS - 1);
        } else if ((i = read(sprp->s, body1, SO1024ANSS - 1)) < 0) {
            ifBLIN_QV1(sprp->flags) warn("read #3");
            ex = EX_IOERR;
        } else if (sprp->asld >= 0) {
            if  (write(sprp->asld, body1, i) < 0) {
                ifBLIN_QV2(sprp->flags) warn("write answer");
        }   }
        ifBLIN_QV4(sprp->flags) fprintf(stderr, "ANSWER=%s~\n", body1);
    }

    /**************************************************************************************************
     *  
     */
    if  (!ex && (babolo_goword(&sowords, (const u_char**)&body1) != TSO_DAEMON)) {
        ifBLIN_QV1(sprp->flags) warnx("proto #3");
        ex = EX_PROTOCOL;
    }

    /**************************************************************************************************
     *     
     */
    if  (!ex) {
        u_int32_t i;
        int alen;

        alen = strtol(body1, &body1, 0);
        if  (!(i = babolo_goword(&mhwords, (const u_char**)&body1))) body1--;
        ifBLIN_QV5(sprp->flags) fprintf(stderr, "L ANSWER=%d\n", alen);
    }

    /**************************************************************************************************
     *   
     */
    if  (!ex) {
        u_int32_t i;
        float f;
        int j;

        for (; !ex;) {
            i = babolo_goword(&sowords, (const u_char**)&body1);
            switch (i) {
            case 0          :
                ifBLIN_QV5(sprp->flags) fprintf(stderr, " spamooborona1024 end of qualificators\n");
                goto l1;
            case TSO_DAEMON :
                ifBLIN_QV1(sprp->flags) warnx("SODAEMON not in begin");
                ex = EX_PROTOCOL;
                goto l1;
            case TSO_REJECT :
                i = strtol(body1, &body1, 10);
                if  (i) sprp->flags |= SPAMOOBORONA1024_ARJC;
                ifBLIN_QV5(sprp->flags) fprintf(stderr, " spamooborona1024 REJECT=%d\n", i);
                break;
            case TSO_REJTXT : break;
            case TSO_SPAM   :
                j = strtol(body1, &body1, 10);
                if  (j) sprp->flags |= SPAMOOBORONA1024_ASPM;
                ifBLIN_QV5(sprp->flags) fprintf(stderr, " spamooborona1024 SPAM=%d\n", j);
                break;
            case TSO_SPAMTXT: break;
            case TSO_SPAMlvl:
#ifdef strtof
                f = strtof(body1, &body1);
#else
                f = atof(body1);
#endif
                if  (  (sprp->flags & SPAMOOBORONA1024_QSPL)
                    && !isnanf(sprp->spamlevel)
                    && (f > sprp->spamlevel)
                    ) {
                    sprp->flags |= SPAMOOBORONA1024_ASPL;
                    ifBLIN_QV5(sprp->flags) fprintf(stderr, " spamooborona1024 spam %f over\n", f);
                } else {
                    ifBLIN_QV5(sprp->flags) fprintf(stderr, " spamooborona1024 SPAM LV=%f\n", f);
                }
                break;
            }
            if  (!babolo_goword(&mhwords, (const u_char**)&body1)) body1--;
    }   }
l1: ;
    /**************************************************************************************************
     * ,     
     */
    if  (!ex) {
        ex = (sprp->flags >> SPAMOOBORONA1024_QSFT)
           & (sprp->flags >> SPAMOOBORONA1024_ASFT)
           & SPAMOOBORONA1024_BMSK;
    }
    /**************************************************************************************************
     *      
     */
    if  (!ex) {
        if  ((ex = so1024_crlf(sprp, (u_char*)maildata, mailsize, 1)) < 0) {
            ex = -ex;
        } else {
            hlen = ex;
            ex = 0;
    }   }

    /**************************************************************************************************
     *      
     */
    if  (!ex) {
        if  ((ex = so1024_crlf(sprp, (u_char*)body1, strlen(body1), 0)) < 0) ex = -ex; else ex = 0;
    }

    /**************************************************************************************************
     *     
     */
    if  (!ex) {
        if  ((ex = fprintf(sprp->out, "\r\n")) < 0) ex = EX_IOERR; else ex = 0;
    }
    if  (!ex) {
        ssize_t len = mailsize - hlen;
        char *data = maildata + hlen;
        off_t offset = hlen;

        while (!ex && (len > 0)) {
            if  ((ex = so1024_crlf(sprp, (u_char*)data, len, 0)) < 0)  ex = -ex; else ex = 0;
            if  (ex) break;
            offset += len;
            if  ((len = mife_read(sprp->md, sprp->maxmailsize, offset)) < 0) {
                ifBLIN_QV1(sprp->flags) warn("spamooborona1024_query: mife_read");
                ex = EX_IOERR;
                break;
            }
            if  (sprp->md->flag & MIFE_EOFL) break;
            if  (!(data = mife_get(sprp->md, offset))) {
                ifBLIN_QV1(sprp->flags) warn("spamooborona1024_query: mife_get");
                ex = EX_IOERR;
                break;
    }   }   }
    ifBLIN_QV4(sprp->flags) fprintf(stderr, "-spamooborona1024_query %d\n", ex);
    return(ex);
}

void
usage(int ex) {
    fprintf( stderr
           , "spamooborona1024 client " VERS "  " DATE " @BABOLO http://www.babolo.ru/\n"
             "|  For Yandex spamooborona details please visit\n"
             "|  http://so.yandex.ru/\n"
             " -d arg - dummy\n"
             " -E - give arg of next flag from environment variable\n"
             " -f file - file with mail\n"
             " -F [r][s] - refuse letter because of:\n"
             "    r - REJECT mark from spamooborona\n"
             "    s - SPAM mark from spamooborona\n"
             " -h host - IP or FQDN of spamooborona server\n"
             " -l float - max acceptable spam level\n"
             " -m - translate verbosea level to mife\n"
             " -o file - file to collect spamooborona answers\n"
             " -p port - port of spamooborona service\n"
             " -q - quiet\n"
             " -r flags - recursive flags without variable substitution, for use in #! string\n"
             " -R flags - recursive flags with variable substitution, for use in #! string\n"
             " -s int - max mail size in bytes\n"
             " -t - test, do not touch spamooborona server\n"
             " -v[v...] - be more verbose\n"
             " -? - short help\n"
             " Example: -E -f EMAIL results in filtering of email from filename ${EMAIL}\n"
           );
    exit(ex);
}

char *
ee(spamooborona1024_root *sprp, char *txt) {
    if  (sprp->flags & SO1024_EFLG) {
        txt = getenv(txt);
    }
    sprp->flags &= ~SO1024_EFLG;
    return(txt);
}

enum classes {n0, nl, nr, no, nx};
/* n0 - \0
   nl - \n
   nr -  
   no -  
   nx
 */

static const enum classes class[256] =
{ n0,no,no,no ,no,no,no,no ,no,nr,nl,nl ,no,nl,no,no
, no,no,no,no ,no,no,no,no ,no,no,no,no ,no,no,no,no
, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr
, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr

, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr
, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr
, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr
, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,no

, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr
, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr
, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr
, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr

, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr
, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr
, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr
, nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr ,nr,nr,nr,nr
};

#define C_st 0x00100000 /*                                                                    */
#define C_ad 0x00080000 /*                                                                */
#define C_el 0x00040000 /*                                                                */
#define C_zr 0x00020000 /*                                                       */
#define C_er 0x00010000 /*                                                                     */
//#define C_state  0xFF /*                                                         */

enum states {d0, d1, dx};
/* d0 -   
   d1 -   
   dx -  
 */

static const u_int32_t automa[nx][dx] =
/*           d0            d1  */
{{ C_el|     dx, C_ad|     dx}/* n0 */
,{ C_el|     d0, C_ad|C_zr|d0}/* nl */
,{ C_st|     d1,           d1}/* nr */
,{ C_el|C_er|dx, C_el|C_er|dx}/* no */
};

#if DEBUG
static const char *clasnames[] = {"n0", "nl", "nr", "no", "nx"};
static const char *pn[] = 
{ "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
, "- ", "- ", "- ", "st", "ad", "el", "zr", "er"
, "- ", "- ", "- ", "- ", "- ", "- ", "- ", "- "
};
static const char *statenames[] = {"d0", "d1", "dx"};
#endif /* DEBUG */

#ifndef NAN
static u_int32_t NaN = 0x7FC00000;
#endif

int
main(int argc, char **argv, char **envp) {
    char *(rcpto[129]), *recipients, *infn;
    u_int32_t mflags = 0;
    spamooborona1024_root spr;
    babolo_opt *opt = NULL;
    int c, ex = EX_OK;
    ssize_t mailsize;
    void *maildata;

    bzero(&spr, sizeof(spr));
    spr.flags = BLIN_VER1 | SPAMOOBORONA1024_RDNS;
    spr.maxmailsize = SO1024MXMS;
#ifdef NAN
    spr.spamlevel = NAN;
#else
    spr.spamlevel = *((float*)&NaN);
#endif
    spr.out = stdout;
    spr.asld = -1;
    spr.s = -1;
    mflags = MIFE_FULL;
    if  (babolo_openopt(Bpars_NONU | BLIN_VER1, &opt, argc, argv, "d:f:F:h:l:mo:p:qr:R:s:tv?")) {
        errx(EX_SOFTWARE, "openopt");
    }
    while ((c = babolo_getopt(&opt)) != 0) {
        babolo_parm *bp;
        char *o, *f;

        switch (c) {
        case 'd': spr.flags &= ~SO1024_EFLG;                                               break;
        case 'E': spr.flags |= SO1024_EFLG;                                                break;
        case 'f': infn = ee(&spr, babolo_getoptarg(&opt, 0));                              break;
        case 'F': for (f = ee(&spr, babolo_getoptarg(&opt, 0)); *f; f++) {
                ;     switch (*f & 0x20) {
                ;     case 'r': spr.flags |= SPAMOOBORONA1024_QRJC; break;
                ;     case 's': spr.flags |= SPAMOOBORONA1024_QSPM; break;
                ; }   }
                ; break;
        case 'h': spr.host = ee(&spr, babolo_getoptarg(&opt, 0));                          break;
        case 'l':
#ifdef strtof
                ; spr.spamlevel = strtof(ee(&spr, babolo_getoptarg(&opt, 0)), NULL);
#else
                ; spr.spamlevel = atof(ee(&spr, babolo_getoptarg(&opt, 0)));
#endif
                ; spr.flags |= SPAMOOBORONA1024_QSPL;
                ; break;
        case 'm': mflags = (spr.flags & BLIN_VERMASK) | MIFE_FULL;
                ; spr.flags &= ~SO1024_EFLG;
                ; break;
        case 'o': spr.asld = open( ee(&spr, babolo_getoptarg(&opt, 0))
                                 , O_WRONLY | O_APPEND | O_CREAT
                                 , 0600
                                 );
                ; break;
        case 'p': spr.port = ee(&spr, babolo_getoptarg(&opt, 0));                          break;
        case 'q': spr.flags &= ~(BLIN_VERMASK | SO1024_EFLG);                              break;
        case 'r': if  (!(o = ee(&spr, babolo_getoptarg(&opt, 0)))) usage(EX_USAGE);
                ; bp = babolo_getparm( Bpars_NOEN | Bpars_NOAL | Bpars_NONU
                                     , (u_char**)&o
                                     , (u_char*)""
                                     , 0
                                     );
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 'R': if  (!(o = ee(&spr, babolo_getoptarg(&opt, 0)))) usage(EX_USAGE);
                ; bp = babolo_getparm(Bpars_NOAL | Bpars_NONU, (u_char**)&o, (u_char*)"", 0);
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 's': spr.maxmailsize = strtoul(ee(&spr, babolo_getoptarg(&opt, 0)), NULL, 0); break;
        case 't': spr.flags |= SPAMOOBORONA1024_TEST;
                ; spr.flags &= ~SO1024_EFLG;
                ; break;
        case 'v': BLIN_VERBOSE(spr.flags);
                ; spr.flags &= ~SO1024_EFLG;
                ; break;
        case '?': usage(EX_OK);
                ; spr.flags &= ~SO1024_EFLG;
                ; break;
        default : usage(EX_USAGE);
    }   }
    infn = babolo_getarg(&opt);
    recipients = getenv("RECIPIENTS");
    if  (!recipients) {
        rcpto[0] = NULL;
    } else {
        u_int32_t state, control;
        char **r;

        state = d0;
        ifBLIN_QV5(spr.flags) BLIN_STATEHEADER(stderr, pn)
        for (r = rcpto; state < dx; recipients++) {
            control = automa[class[*recipients & 0xFF]][state];
            ifBLIN_QV5(spr.flags)
                BLIN_STATEBODY( stderr\
                              , pn\
                              , 0xFF\
                              , clasnames[class[*recipients & 0xFF]]\
                              , recipients\
                              , 0\
                              , BLIN_I(recipients)\
                              , control\
                              , statenames\
                              , state\
                              , C_state\
                              , (recipients[BLIN_STATEVAR_i]!='\n')\
                              )
            if  (control & C_st) *r = recipients;
            if  (control & C_ad) r++;
            if  (control & (C_ad|C_el)) *r = NULL;
            if  (control & C_zr) *recipients = 0;
            if  (control & C_er) ex = EX_DATAERR;
            state = control & C_state;
    }   }
    if  (!ex) {
        if  (infn) spr.md = mife_open(mflags, infn); else spr.md = mife_opef(mflags, fileno(stdin));
        if  (!spr.md) {
            ifBLIN_QV1(spr.flags) warn("mife_open");
            ex = EX_IOERR;
    }   }
    if  (!ex && ((mailsize = mife_read(spr.md, spr.maxmailsize, 0)) < 0)) {
        ifBLIN_QV1(spr.flags) warn("mife_read");
        ex = EX_IOERR;
    }
    if  (!ex && !(maildata = mife_get(spr.md, 0))) {
        ifBLIN_QV1(spr.flags) warn("mife_get");
        ex = EX_IOERR;
    }
    if  (!ex) ex = spamooborona1024_query( &spr
                                         , getenv("CLIENT")
                                         , NULL
                                         , NULL
                                         , getenv("SENDER")
                                         , rcpto
                                         , mailsize
                                         , maildata
                                         );
    if  (spr.asld >= 0) close(spr.asld);
    exit(ex);
}
